import typing

from ..liteclient.client import LiteClient
from ..boc.cell import Cell
from ..boc.address import Address
from ..tlb.account import StateInit, Account, SimpleAccount, ShardAccount
from ..tlb.block import CurrencyCollection
from ..tlb.transaction import ExternalMsgInfo, MessageAny, InternalMsgInfo


class ContractError(BaseException):
    pass


class Contract:

    def __init__(self, provider: LiteClient, address: Address, account: typing.Optional[Account] = None,
                 shard_account: typing.Optional[ShardAccount] = None, state_init: typing.Optional[StateInit] = None,
                 **kwargs):
        """
        :param provider: LiteClient
        :param address: contract address
        :param account: full account state
        :param shard_account: shard account
        :param state_init: account state init. usually used for contract deploying
        :param kwargs: some additional account attributes. for e.g. private key for wallet contracts
        """
        self.provider = provider
        self.address = address
        self.raw_account: typing.Optional[Account] = None
        self.account: SimpleAccount = None
        self.shard_account: typing.Optional[ShardAccount] = None
        self.is_active: bool = None
        self.is_uninitialized: bool = None
        self.is_frozen: bool = None
        self.state: typing.Optional[StateInit] = None
        self.state_init: typing.Optional[StateInit] = state_init
        self.set_account_attributes(account, shard_account)

        for k, v in kwargs.items():
            setattr(self, k, v)

    def set_account_attributes(self, account: typing.Optional[Account], shard_account: typing.Optional[ShardAccount]):
        self.raw_account = account
        self.account = SimpleAccount.from_raw(account, self.address)
        self.shard_account = shard_account
        self.is_active = self.account.is_active()
        self.is_uninitialized = self.account.is_uninitialized()
        self.is_frozen = self.account.is_frozen()

        self.state = None
        if self.is_active:
            self.state = self.account.state.state_init
        else:
            self.state = self.state_init

    @property
    def data(self) -> Cell:
        return self.state.data

    @property
    def code(self) -> Cell:
        return self.state.code

    @property
    def balance(self) -> int:
        """
        :return: balance from saved account state
        """
        return self.account.balance

    @classmethod
    async def from_address(cls, provider: LiteClient, address: typing.Union[str, Address], **kwargs):
        if isinstance(address, str):
            address = Address(address)
        account, shard_account = await provider.raw_get_account_state(address)
        return cls(provider=provider, address=address, account=account, shard_account=shard_account, **kwargs)

    @classmethod
    async def from_state_init(cls, provider: LiteClient, workchain: int, state_init: StateInit, **kwargs):
        address = Address((workchain, state_init.serialize().hash))
        return await cls.from_address(provider=provider, address=address, state_init=state_init, **kwargs)

    @classmethod
    async def from_code_and_data(cls, provider: LiteClient, workchain: int, code: Cell, data: Cell, **kwargs):
        state_init = StateInit(code=code, data=data)
        return await cls.from_state_init(provider=provider, workchain=workchain, state_init=state_init, **kwargs)

    async def update(self):
        account, shard_account = await self.raw_get_account_state()
        self.set_account_attributes(account, shard_account)

    async def raw_get_account_state(self):
        return (await self.provider.raw_get_account_state(address=self.address))[0]

    async def get_account_state(self):
        return await self.provider.get_account_state(self.address)

    async def get_balance(self) -> int:
        """
        :return: balance from current account state. better to
            await contract.update()
            contract.balance
        """
        account_state = await self.get_account_state()
        return account_state.balance

    async def run_get_method(self, method: typing.Union[str, int], stack: typing.Optional[list] = None):
        if stack is None:
            stack = []
        return await self.provider.run_get_method(self.address, method, stack)

    @staticmethod
    def create_external_msg(src: typing.Optional[Address] = None, dest: typing.Optional[Address] = None,
                            import_fee: int = 0, state_init: typing.Optional[StateInit] = None,
                            body: Cell = None) -> MessageAny:
        info = ExternalMsgInfo(src, dest, import_fee)
        if body is None:
            body = Cell.empty()
        message = MessageAny(info=info, init=state_init, body=body)
        return message

    @staticmethod
    def create_internal_msg(ihr_disabled: bool = True, bounce: bool = None, bounced: bool = False, src: Address = None,
                            dest: Address = None,
                            value: typing.Union[CurrencyCollection, int] = 0, ihr_fee: int = 0, fwd_fee: int = 0,
                            created_lt: int = 0,
                            created_at: int = 0, state_init: typing.Optional[StateInit] = None,
                            body: Cell = None) -> MessageAny:
        if isinstance(value, int):
            value = CurrencyCollection(grams=value, other=None)
        if bounce is None:
            bounce = dest.is_bounceable
        info = InternalMsgInfo(ihr_disabled, bounce, bounced, src, dest, value, ihr_fee, fwd_fee, created_lt, created_at)
        if body is None:
            body = Cell.empty()
        message = MessageAny(info=info, init=state_init, body=body)
        return message

    async def send_external(self, src: typing.Optional[Address] = None, import_fee: int = 0,
                            state_init: typing.Optional[StateInit] = None, body: Cell = None):
        message = self.create_external_msg(src=src, dest=self.address, import_fee=import_fee, state_init=state_init,
                                           body=body)
        return await self.provider.raw_send_message(message.serialize().to_boc())

    async def send_init_external(self):
        if not self.state_init:
            raise ContractError('contract does not have state_init attribute')
        return await self.send_external(state_init=self.state_init)

    async def deploy_via_external(self):
        return await self.send_init_external()

    def __repr__(self):
        return f'<{self.account.state.type_} {self.__class__.__name__} {self.address}>'  # <active WalletV4R2 EQBvW8Z5huBkMJYdnfAEM5JqTNkuWX3diqYENkWsIL0XggGG>



import typing

from ..tlb import TlbScheme
from ...boc import Cell, Builder, Slice, HashMap, Address


class NftItemData(TlbScheme):
    """
    nft_item_data#_ index:uint64 collection_address:Address owner_address:Address content:^Cell = NftItemData;
    """
    def __init__(self,
                 index: typing.Optional[int] = 0,
                 collection_address: typing.Optional[Address] = None,
                 owner_address: typing.Optional[Address] = None,
                 content: typing.Optional[Cell] = None
                 ):
        self.index = index
        if isinstance(collection_address, str):
            collection_address = Address(collection_address)
        if isinstance(owner_address, str):
            collection_address = Address(collection_address)
        self.collection_address = collection_address
        self.owner_address = owner_address
        self.content = content

    def serialize(self) -> Cell:
        builder = Builder()
        builder\
            .store_uint(self.index, 64)\
            .store_address(self.collection_address)\
            .store_address(self.owner_address)\
            .store_ref(self.content)
        return builder.end_cell()

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        return cls(index=cell_slice.load_uint(64), collection_address=cell_slice.load_address(), owner_address=cell_slice.load_address(), content=cell_slice.load_ref())





import typing

import hashlib
import x25519

from nacl.signing import SigningKey as ed25519Private, VerifyKey as ed25519Public
from nacl.public import PublicKey as x25519Public, PrivateKey as x25519Private

from Cryptodome.Random import get_random_bytes
from Cryptodome.Cipher import AES


class Server:
    def __init__(self, host: str, port: int, pub_key: bytes, magic: bytes = b'\xc6\xb4\x13\x48'):
        self.host = host
        self.port = port
        self.ed25519_public = ed25519Public(pub_key)
        self.x25519_public = self.ed25519_public.to_curve25519_public_key()
        self.magic = magic

    def get_key_id(self):
        return hashlib.sha256(self.magic + self.ed25519_public.encode()).digest()


class Client:
    def __init__(self, ed25519_private_key: bytes) -> None:
        self.ed25519_private = ed25519Private(ed25519_private_key)
        self.ed25519_public = self.ed25519_private.verify_key
        self.x25519_private = self.ed25519_private.to_curve25519_private_key()
        self.x25519_public = self.x25519_private.public_key

    @staticmethod
    def generate_ed25519_private_key() -> bytes:
        return bytes(ed25519Private.generate())


def get_random(bytes_size: int) -> bytes:
    return get_random_bytes(bytes_size)


def create_aes_ctr_cipher(key: bytes, iv: bytes):
    if len(key) != 32:
        raise Exception('key should be 32 bytes exactly!')

    cipher = AES.new(key, AES.MODE_CTR, initial_value=iv, nonce=b'')
    # cipher = Cipher(algorithms.AES(key), modes.CTR(iv)).encryptor()

    return cipher


def aes_ctr_encrypt(cipher, data: bytes) -> bytes:
    # return cipher.encryptor().update(data)
    return cipher.encrypt(data)


def aes_ctr_decrypt(cipher, data: bytes) -> bytes:
    # return cipher.decryptor().update(data)
    return cipher.decrypt(data)


def get_shared_key(private_key: bytes, public_key: bytes) -> bytes:
    """
    :param public_key: peer public x25519 key
    :param private_key: client private x25519 key
    :return: ECDH x25519 shared key
    """
    return x25519.scalar_mult(private_key, public_key)



from abc import ABC, abstractmethod


class TlbError(BaseException):
    pass


class TlbScheme(ABC):
    """
    abstract class for Tlb Schemes wrappers
    """
    @abstractmethod
    def serialize(self, *args): ...

    @classmethod
    @abstractmethod
    def deserialize(cls, *args): ...

    def __repr__(self):
        return str(self.__dict__)
        # TODO beautiful repr
